from .colormaps import (
    matplotlib_colormaps,
    simple_colormaps,
    ensure_colormap_tuple,
    AVAILABLE_COLORMAPS,
    ALL_COLORMAPS,
    MAGENTA_GREEN,
    RGB,
    CYMRGB,
)
from .colorbars import make_colorbar
